@@ -1,5 +1,5 @@
 /* ----------------------------------------------------------------------- *
- *   
+ *
  *   Copyright 2001-2006 H. Peter Anvin - All Rights Reserved
  *
  *   This program is free software available under the same license
@@ -19,5 +19,4 @@
 
 int
 myrecvfrom(int s, void *buf, int len, unsigned int flags,
-	   struct sockaddr *from, socklen_t *fromlen,
-	   struct sockaddr_in *myaddr);
+           union sock_addr *from, union sock_addr *myaddr);
