@@ -1,7 +1,6 @@
-/* $Id: remap.h,v 1.6 2002/11/08 01:10:09 hpa Exp $ */
 /* ----------------------------------------------------------------------- *
- *   
- *   Copyright 2001 H. Peter Anvin - All Rights Reserved
+ *
+ *   Copyright 2001-2007 H. Peter Anvin - All Rights Reserved
  *
  *   This program is free software available under the same license
  *   as the "OpenBSD" operating system, distributed at
@@ -27,7 +26,7 @@ struct rule;
    macro character is passed as the first argument; the output buffer,
    if any, is passed as the second argument.  The function should return
    the number of characters output, or -1 on failure. */
-typedef int (*match_pattern_callback)(char, char *);
+typedef int (*match_pattern_callback) (char, char *);
 
 /* Read a rule file */
 struct rule *parserulefile(FILE *);
@@ -36,9 +35,8 @@ struct rule *parserulefile(FILE *);
 void freerules(struct rule *);
 
 /* Execute a rule set on a string; returns a malloc'd new string. */
-char *rewrite_string(const char *, const struct rule *, int,
-		     match_pattern_callback, const char **);
-
-#endif /* WITH_REGEX */
-#endif /* TFTPD_REMAP_H */
+char *rewrite_string(const char *, const struct rule *, char,
+                     match_pattern_callback, const char **);
 
+#endif                          /* WITH_REGEX */
+#endif                          /* TFTPD_REMAP_H */
