@@ -1,6 +1,5 @@
-/* $Id$ */
 /* ----------------------------------------------------------------------- *
- *   
+ *
  *   Copyright 2001 H. Peter Anvin - All Rights Reserved
  *
  *   This program is free software available under the same license
@@ -24,4 +23,13 @@ char *tfstrdup(const char *);
 
 extern int verbosity;
 
+struct formats {
+    const char *f_mode;
+    char *(*f_rewrite) (const struct formats *, char *, int, int, const char **);
+    int (*f_validate) (char *, int, const struct formats *, const char **);
+    void (*f_send) (const struct formats *, struct tftphdr *, int);
+    void (*f_recv) (const struct formats *, struct tftphdr *, int);
+    int f_convert;
+};
+
 #endif
